import numpy as np
import sklearn.utils as utils
import math
from matplotlib.patches import Ellipse
import matplotlib.transforms as transforms

# sample = [1xM], clusters = [KxM]
def distance(sample,clusters):
    sample_rep = np.matlib.repmat(sample,len(clusters[:,1]),1)
    dist = clusters - sample_rep
    # return size [1xK]
    return np.linalg.norm(dist,2,axis=1)**2

# sample = [1xM], clusters = [KxM]
def min_distance(sample,clusters):
    dist = distance(sample,clusters)
    # return scalar
    return np.min(dist)

# sample = [1xM], clusters = [KxM]
def argmin_distance(sample,clusters):
    dist = distance(sample,clusters)
    # return scalar
    return np.argmin(dist)

# samples = [NxM], clusters = [KxM]
def mean_distance(samples,clusters):
    dist = np.zeros(len(samples[:,1]))
    
    for i in range(len(samples[:,1])):
        dist[i] = min_distance(samples[i,:],clusters)
    
    # return scalar
    return np.mean(dist)

# samples = [MxN], n_clusters = K, max_iterations = positive scalar, clusters_init = [K,N] (optional)
def kmeans_custom(samples, n_clusters, max_iterations, clusters_init = None):
    
    # Some termination thresholds
    dist_thr = 1/n_clusters
    stab_thr = 1 - 1*10**-10
    
    # Split into training and test set (e.g. by shuffling or by generating random indices)
    samples_shuffled = np.copy(samples)
    np.random.shuffle(samples_shuffled)
    
    # Split with fixed ratio
    samples_train = samples_shuffled[0:int(0.8*len(samples[:,1])),:]
    samples_eval = samples_shuffled[int(0.8*len(samples[:,1]))+1:len(samples[:,1]),:]
    
    if n_clusters > len(samples_train[:,1]):
        print('Error: number of clusters larger than number of available training vectors!')
        return None
    
    # Cluster initilization
    if clusters_init is not None:        
        clusters = np.copy(clusters_init)
    else:       
        init = utils.random.sample_without_replacement(n_population=len(samples_train[:,1]), n_samples=n_clusters)
        clusters = samples_train[init,:]
    
    # Allocate
    dist_vec = np.zeros([len(samples_train[:,1])],dtype='float')
    class_vec = np.zeros([len(samples_train[:,1])],dtype='int')
    
    # Mean distance before correction
    dist_init = mean_distance(samples_eval,clusters)    
    
    # Notification
    print('Mean distance gain [dB]:')

    # Correction
    for i in range(max_iterations):        
        dist_prev = mean_distance(samples_eval,clusters)
        clusters_prev = clusters
        
        for j in range(len(samples_train[:,1])):            
            dist_vec[j] = min_distance(samples_train[j,:],clusters)
            class_vec[j] = argmin_distance(samples_train[j,:],clusters)        
        
        for c in range(n_clusters):
            candidate = np.mean(samples_train[class_vec==c], axis=0)
            
            # Why is this necessary?
            if np.any(np.isnan(candidate)):
                clusters[c,:] = clusters[c,:]
            else:
                clusters[c,:] = candidate      
        
        dist_cur = mean_distance(samples_eval,clusters)
        print(str(i+1) + ': ' + str(20*math.log10(dist_cur/dist_init)))
        
        # Termination before maximum number of iterations
        if dist_cur < dist_thr:            
            print('\nTerminate K-means due to achieving distance threshold. Mean distance is {}.\n'.format(dist_cur))
            return clusters, dist_cur
        
        elif dist_cur/dist_prev > stab_thr:         
            print('\nTerminate K-means due to convergence/stability criterion. Mean distance is {}.\n'.format(dist_prev))
            return clusters_prev, dist_prev
            
    print('\nTerminate K-means due to maximum number of iterations. Mean distance is {}.\n'.format(dist_cur))        
    return clusters, dist_cur

# Covariance indicator
def elips(var,cluster,ax,color):
    vals, vecs = np.linalg.eigh(var)
    order = np.argsort(vals)[::-1]
    vals = vals[order]
    vecs = vecs[:,order]
    
    theta = np.degrees(np.arctan2(vecs[1,0],vecs[0,0]))
    ell = Ellipse(xy=(0,0),width=2*np.sqrt(vals[0]),height=2*np.sqrt(vals[1]),\
                  angle = theta,edgecolor=color,fill=False)
    
    transf = transforms.Affine2D().translate(cluster[0,0],cluster[0,1])
    ell.set_transform(transf + ax.transData)
    ax.add_patch(ell)